import boto3
import logging
import os
import tempfile

from botocore.exceptions import ClientError, EndpointConnectionError, ConnectionClosedError
from io import BytesIO

from assemblyline.common.exceptions import ChainAll
from assemblyline.filestore.transport.base import Transport, TransportException

try:
    from botocore.vendored.requests.packages.urllib3 import disable_warnings
except ImportError:
    from urllib3 import disable_warnings


disable_warnings()

"""
This class assumes a flat file structure in the S3 bucket.  This is due to the way the AL datastore currently handles
file paths for local/ftp datastores not playing nicely with s3 constraints.
"""


@ChainAll(TransportException)
class TransportS3(Transport):
    DEFAULT_HOST = "s3.amazonaws.com"

    def __init__(self, base=None, accesskey=None, secretkey=None, aws_region=None, s3_bucket="al-storage",
                 host=None, port=None, use_ssl=None, verify=True, connection_attempts=None):
        self.log = logging.getLogger('assemblyline.transport.s3')
        self.base = base
        self.bucket = s3_bucket
        self.accesskey = accesskey
        self.retry_limit: int = connection_attempts

        if use_ssl is None:
            self.use_ssl = True
        else:
            self.use_ssl = use_ssl

        if host is None:
            self.host = self.DEFAULT_HOST
        else:
            self.host = host

        if port is None:
            self.port = {True: 443, False: 80}[self.use_ssl]
        else:
            self.port = port

        self.scheme = {True: "https", False: "http"}[self.use_ssl]

        self.endpoint_url = "{scheme}://{host}:{port}".format(scheme=self.scheme, host=self.host, port=self.port)

        session = boto3.session.Session()
        self.client = session.client(
            "s3",
            aws_access_key_id=accesskey,
            aws_secret_access_key=secretkey,
            endpoint_url=self.endpoint_url,
            region_name=aws_region,
            use_ssl=self.use_ssl,
            verify=verify
        )

        bucket_exist = False
        try:
            self.with_retries(self.client.head_bucket, Bucket=self.bucket)
            bucket_exist = True
        except TransportException as e:
            if isinstance(e.cause, ClientError) and e.cause.response['Error']['Code'] == '404':
                pass
            else:
                raise

        if not bucket_exist:
            try:
                self.with_retries(self.client.create_bucket, Bucket=self.bucket)
            except TransportException as e:
                if isinstance(e.cause, ClientError) and e.cause.response['Error']['Code'] == "BucketAlreadyOwnedByYou":
                    # By the time that we listed the bucket an found it didn't exist, someone else created it.
                    pass
                else:
                    raise

        def s3_normalize(path):
            # flatten path to just the basename
            return os.path.basename(path)

        super(TransportS3, self).__init__(normalize=s3_normalize)

    def __str__(self):
        out = "s3://"
        if self.accesskey:
            out += "%s@" % self.accesskey
        out += "%s:%s" % (self.host, self.port)
        if self.bucket:
            out += "/%s" % self.bucket
        if self.base:
            out += self.base
        return out

    def with_retries(self, func, *args, **kwargs):
        retries = 0
        while self.retry_limit is None or retries <= self.retry_limit:
            try:
                ret_val = func(*args, **kwargs)

                if retries:
                    self.log.info('Reconnected to S3 transport!')

                return ret_val

            except (EndpointConnectionError, ConnectionClosedError):
                self.log.warning(f"No connection to S3 transport {self.endpoint_url}, retrying...")
                retries += 1
        raise ConnectionError(f"Couldn't connect to the requested S3 endpoint {self.endpoint_url} inside retry limit")

    def delete(self, path):
        key = self.normalize(path)
        self.with_retries(self.client.delete_object, Bucket=self.bucket, Key=key)

    def exists(self, path):
        # checks to see if KEY exists
        key = self.normalize(path)
        self.log.debug('Checking for existence of %s', key)
        try:
            self.with_retries(self.client.head_object, Bucket=self.bucket, Key=key)
        except TransportException as error:
            if isinstance(error.cause, ClientError):
                return False
            raise
        return True

    def makedirs(self, path):
        # Does not need to do anything as s3 has a flat layout.
        pass

    # File based functions
    def download(self, src_path, dst_path):
        key = self.normalize(src_path)
        dir_path = os.path.dirname(dst_path)
        # create dst_path if it doesn't exist
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        # download the key from s3
        self.with_retries(self.client.download_file, self.bucket, key, dst_path)

    def upload(self, src_path, dst_path):
        dst_path = self.normalize(dst_path)
        # if file exists already, it will be overwritten
        self.with_retries(self.client.upload_file, src_path, self.bucket, dst_path)

    # Buffer based functions
    def get(self, path: str) -> bytes:
        fd, dst_path = tempfile.mkstemp(prefix="s3_transport.", suffix=".download")
        os.close(fd)  # We don't need the file descriptor open

        self.with_retries(self.download, path, dst_path)
        try:
            with open(dst_path, "rb") as downloaded:
                return downloaded.read()
        finally:
            if os.path.exists(dst_path):
                os.remove(dst_path)

    def put(self, dst_path, content):
        dst_path = self.normalize(dst_path)
        if isinstance(content, str):
            content = content.encode('utf-8')

        with BytesIO(content) as file_io:
            self.with_retries(self.client.upload_fileobj, file_io, self.bucket, dst_path)



from __future__ import annotations
from typing import Optional as Opt

from assemblyline import odm
from assemblyline.common import forge
from assemblyline.common.constants import DEFAULT_SERVICE_ACCEPTS, DEFAULT_SERVICE_REJECTS

Classification = forge.get_classification()

SIGNATURE_DELIMITERS = {
    'new_line': '\n',
    'double_new_line': '\n\n',
    'pipe': '|',
    'comma': ',',
    'space': ' ',
    'none': '',
    'file': '',
    'custom': ''
}


@odm.model(index=False, store=False)
class EnvironmentVariable(odm.Model):
    name: str = odm.Keyword()
    value: str = odm.Keyword()


@odm.model(index=False, store=False)
class DockerConfig(odm.Model):
    allow_internet_access: bool = odm.Boolean(default=False)
    command: Opt[list[str]] = odm.Optional(odm.List(odm.Keyword()))
    cpu_cores: float = odm.Float(default=1.0)
    environment: list[EnvironmentVariable] = odm.List(odm.Compound(EnvironmentVariable), default=[])
    image: str = odm.Keyword()                       # Complete name of the Docker image with tag, may include registry
    registry_username: Opt[str] = odm.Optional(odm.Keyword())  # The username to use when pulling the image
    registry_password: Opt[str] = odm.Optional(odm.Keyword())  # The password or token to use when pulling the image
    registry_type: str = odm.Enum(
        values=["docker", "harbor"],
        default='docker')  # The type of registry (Docker, Harbor)
    ports: list[str] = odm.List(odm.Keyword(), default=[])
    ram_mb: int = odm.Integer(default=512)
    ram_mb_min: int = odm.Integer(default=128)


@odm.model(index=False, store=False)
class PersistentVolume(odm.Model):
    mount_path = odm.Keyword()  # Path into the container to mount volume
    capacity = odm.Keyword()  # Bytes
    storage_class = odm.Keyword()


@odm.model(index=False, store=False)
class DependencyConfig(odm.Model):
    container: DockerConfig = odm.Compound(DockerConfig)
    volumes = odm.Mapping(odm.Compound(PersistentVolume), default={})
    run_as_core: bool = odm.Boolean(default=False)


@odm.model(index=False, store=False)
class UpdateSource(odm.Model):
    name: str = odm.Keyword()
    password: Opt[str] = odm.Optional(odm.Keyword(default=""))
    pattern: Opt[str] = odm.Optional(odm.Keyword(default=""))
    private_key: Opt[str] = odm.Optional(odm.Keyword(default=""))
    ca_cert: Opt[str] = odm.Optional(odm.Keyword(default=""))
    ssl_ignore_errors: bool = odm.Boolean(default=False)
    proxy: Opt[str] = odm.Optional(odm.Keyword(default=""))
    uri: str = odm.Keyword()
    username: Opt[str] = odm.Optional(odm.Keyword(default=""))
    headers: list[EnvironmentVariable] = odm.List(odm.Compound(EnvironmentVariable), default=[])
    default_classification = odm.Classification(default=Classification.UNRESTRICTED)


@odm.model(index=False, store=False)
class UpdateConfig(odm.Model):
    generates_signatures = odm.Boolean(index=True, default=False)
    sources = odm.List(odm.Compound(UpdateSource), default=[])    # Generic external resources we need
    update_interval_seconds = odm.Integer()                       # Update check interval in seconds
    wait_for_update = odm.Boolean(default=False)
    signature_delimiter = odm.Enum(values=SIGNATURE_DELIMITERS.keys(),
                                   default="double_new_line")           # Delimiter use in between signature
    custom_delimiter = odm.Optional(odm.Keyword())                      # Custom delimiter


@odm.model(index=False, store=False)
class SubmissionParams(odm.Model):
    default = odm.Any()
    name = odm.Keyword()
    type = odm.Enum(values=['str', 'int', 'list', 'bool'])
    value = odm.Any()
    list = odm.Optional(odm.Any())
    hide = odm.Boolean(default=False)


@odm.model(index=True, store=False)
class Service(odm.Model):
    # Regexes applied to assemblyline style file type string
    accepts = odm.Keyword(store=True, default=DEFAULT_SERVICE_ACCEPTS)
    rejects = odm.Optional(odm.Keyword(store=True, default=DEFAULT_SERVICE_REJECTS))

    category = odm.Keyword(store=True, default="Static Analysis", copyto="__text__")
    config = odm.Mapping(odm.Any(), default={}, index=False, store=False)
    description = odm.Text(store=True, default="NA", copyto="__text__")
    default_result_classification = odm.ClassificationString(default=Classification.UNRESTRICTED)
    enabled: bool = odm.Boolean(store=True, default=False)
    is_external: bool = odm.Boolean(default=False)
    licence_count: int = odm.Integer(default=0)

    name: str = odm.Keyword(store=True, copyto="__text__")
    version = odm.Keyword(store=True)

    # Should the result cache be disabled for this service
    disable_cache = odm.Boolean(default=False)

    stage = odm.Keyword(store=True, default="CORE", copyto="__text__")
    submission_params: SubmissionParams = odm.List(odm.Compound(SubmissionParams), index=False, default=[])
    timeout: int = odm.Integer(default=60)

    docker_config: DockerConfig = odm.Compound(DockerConfig)
    dependencies: dict[str, DependencyConfig] = odm.Mapping(odm.Compound(DependencyConfig), default={})

    update_channel: str = odm.Enum(values=["stable", "rc", "beta", "dev"], default='stable')
    update_config: UpdateConfig = odm.Optional(odm.Compound(UpdateConfig))



from __future__ import annotations
import hashlib
from assemblyline import odm
from assemblyline.common import forge
Classification = forge.get_classification()

SUBMISSION_STATES = ['failed', 'submitted', 'completed']
DEFAULT_SRV_SEL = ["Filtering", "Antivirus", "Static Analysis", "Extraction", "Networking"]
DEFAULT_RESUBMIT = []


@odm.model(index=True, store=False)
class File(odm.Model):
    name = odm.Keyword(copyto="__text__")    # Name of the file
    size = odm.Optional(odm.Integer())       # Size of the file
    sha256 = odm.SHA256(copyto="__text__")   # SHA256 hash of the file


@odm.model(index=False, store=False)
class ServiceSelection(odm.Model):
    selected = odm.List(odm.Keyword(), default=DEFAULT_SRV_SEL)   # List of selected services
    excluded = odm.List(odm.Keyword(), default=[])                # List of excluded services
    resubmit = odm.List(odm.Keyword(), default=DEFAULT_RESUBMIT)  # Add to service selection when resubmitting
    runtime_excluded = odm.List(odm.Keyword(), default=[])        # List of runtime excluded services


# Fields in the parameters used to calculate hashes used for result caching
_KEY_HASHED_FIELDS = {
    'classification',
    'deep_scan',
    'ignore_cache',
    'ignore_dynamic_recursion_prevention',
    'ignore_filtering',
    'ignore_size',
    'max_extracted',
    'max_supplementary',
}


@odm.model(index=True, store=False)
class SubmissionParams(odm.Model):
    classification = odm.Classification(
        default=Classification.UNRESTRICTED)                            # Original classification of the submission
    deep_scan = odm.Boolean(default=False)                              # Should a deep scan be performed?
    description = odm.Text(store=True, copyto="__text__")               # Description of the submission
    generate_alert = odm.Boolean(default=False)                         # Should this submission generate an alert
    groups = odm.List(odm.Keyword(), default=["USERS"])                 # List of groups related to this scan
    ignore_cache = odm.Boolean(default=False)                           # ignore the service caching or not
    ignore_dynamic_recursion_prevention = odm.Boolean(default=False)    # Should we ignore dynamic recursion prevention
    ignore_filtering = odm.Boolean(default=False)                       # Should we ignore filtering services
    ignore_size = odm.Boolean(default=False)                            # ignore the file size limits
    never_drop = odm.Boolean(default=False)                             # Exempt from being dropped by ingester
    malicious = odm.Boolean(default=False)                              # Is the file submitted known to be malicious
    max_extracted = odm.Integer(default=500)                            # Max number of extracted files
    max_supplementary = odm.Integer(default=500)                        # Max number of supplementary files
    priority = odm.Integer(default=1000)                                # Priority of the scan
    profile = odm.Boolean(default=False)                                # Should the submission do extra profiling
    psid = odm.Optional(odm.UUID())                                     # Parent submission ID
    quota_item = odm.Boolean(default=False)                             # Does this submission count against quota
    services = odm.Compound(ServiceSelection, default={})               # Service selection bloc
    service_spec = odm.Mapping(odm.Mapping(odm.Any()),
                               default={}, index=False, store=False)    # Service specific parameters
    submitter = odm.Keyword(store=True, copyto="__text__")              # User who submitted the file
    ttl = odm.Integer(default=0)                                        # Time to live for this submission in days
    type = odm.Keyword(default="USER")                                  # Type of submission
    initial_data = odm.Optional(odm.Text(index=False))                  # Initialization for auxiliary 'temporary_data'

    def get_hashing_keys(self):
        """Get the sections of the submission parameters that should be used in result hashes."""
        data = self.as_primitives()
        return {k: v for k, v in data.items() if k in _KEY_HASHED_FIELDS}

    def create_filescore_key(self, sha256, services: list = None):
        """This is the key used to store the final score of a submission for fast lookup.

        This lookup is one of the methods used to check for duplication in ingestion process,
        so this key is fairly sensitive.
        """
        # TODO do we need this version thing still be here?
        # One up this if the cache is ever messed up and we
        # need to quickly invalidate all old cache entries.
        version = 0

        if services is None:
            services = self.services.selected

        data = self.get_hashing_keys()
        data['service_spec'] = sorted((key, sorted(values.items())) for key, values in self.service_spec.items())
        data['sha256'] = sha256
        data['services'] = [str(x) for x in services]

        s = ', '.join([f"{k}: {data[k]}" for k in sorted(data.keys())])

        return 'v'.join([str(hashlib.md5(s.encode()).hexdigest()), str(version)])


@odm.model(index=True, store=True)
class Times(odm.Model):
    completed = odm.Optional(odm.Date(store=False))  # Date at which the submission finished scanning
    submitted = odm.Date(default="NOW")              # Date at which the submission started scanning


@odm.model(index=True, store=False)
class Verdict(odm.Model):
    malicious = odm.List(odm.Keyword(), default=[])      # List of user that thinks this submission is malicious
    non_malicious = odm.List(odm.Keyword(), default=[])  # List of user that thinks this submission is non-malicious


@odm.model(index=True, store=True)
class Submission(odm.Model):
    archive_ts = odm.Date(store=False)                          # Archiving timestamp
    classification = odm.Classification()                       # Classification of the submission
    error_count = odm.Integer()                                 # Total number of errors in the submission
    errors: list[str] = odm.List(odm.Keyword(), store=False)    # List of error keys
    expiry_ts = odm.Optional(odm.Date(store=False))             # Expiry timestamp
    file_count = odm.Integer()                                  # Total number of files in the submission
    files: list[File] = odm.List(odm.Compound(File))            # List of files that were originally submitted
    max_score = odm.Integer()                                   # Maximum score of all the files in the scan
    metadata = odm.FlattenedObject(store=False)                 # Metadata associated to the submission
    params: SubmissionParams = odm.Compound(SubmissionParams)   # Submission detail blocs
    results: list[str] = odm.List(odm.Keyword(), store=False)   # List of result keys
    sid = odm.UUID(copyto="__text__")                           # Submission ID
    state = odm.Enum(values=SUBMISSION_STATES)                  # Status of the submission
    times = odm.Compound(Times, default={})                     # Timing bloc
    verdict = odm.Compound(Verdict, default={})                 # Verdict timing

    # the filescore key, used in deduplication. This is a non-unique key, that is
    # shared by submissions that may be processed as duplicates.
    scan_key = odm.Optional(odm.Keyword(store=False, index=False))

    def is_submit(self):
        return self.state == 'submitted'

    def is_complete(self):
        return self.state == 'completed'

    def is_initial(self):
        return self.is_submit() and not self.params.psid



from assemblyline import odm
from . import PerformanceTimer

MSG_TYPES = {"ServiceTimingHeartbeat"}
LOADER_CLASS = "assemblyline.odm.messages.service_heartbeat.ServiceTimingMessage"


@odm.model()
class Metrics(odm.Model):
    execution = PerformanceTimer()
    execution_count = odm.Integer()
    idle = PerformanceTimer()
    idle_count = odm.Integer()


@odm.model()
class Heartbeat(odm.Model):
    instances = odm.Integer()
    metrics = odm.Compound(Metrics)
    queue = odm.Integer()
    service_name = odm.Keyword()


@odm.model()
class ServiceTimingMessage(odm.Model):
    msg = odm.Compound(Heartbeat)
    msg_loader = odm.Enum(values={LOADER_CLASS}, default=LOADER_CLASS)
    msg_type = odm.Enum(values=MSG_TYPES, default="ServiceTimingHeartbeat")
    sender = odm.Keyword()



from assemblyline import odm


# Model definition
@odm.model(index=True, store=False)
class Tagging(odm.Model):
    @odm.model(index=True, store=False)
    class Attribution(odm.Model):
        actor = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        campaign = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        exploit = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        implant = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        family = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        network = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

    @odm.model(index=True, store=False)
    class AV(odm.Model):
        heuristic = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        virus_name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

    @odm.model(index=True, store=False)
    class Cert(odm.Model):
        @odm.model(index=True, store=False)
        class CertValid(odm.Model):
            start = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            end = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

        extended_key_usage = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        issuer = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        key_usage = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        owner = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        serial_no = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        signature_algo = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        subject = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        subject_alt_name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        thumbprint = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        valid = odm.Optional(odm.Compound(CertValid))
        version = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

    @odm.model(index=True, store=False)
    class Dynamic(odm.Model):
        @odm.model(index=True, store=False)
        class DynamicProcess(odm.Model):
            command_line = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            file_name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            shortcut = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

        @odm.model(index=True, store=False)
        class DynamicSignature(odm.Model):
            category = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            family = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

        @odm.model(index=True, store=False)
        class DynamicSSDeep(odm.Model):
            cls_ids = odm.Optional(odm.List(odm.SSDeepHash(copyto="__text__")))
            dynamic_classes = odm.Optional(odm.List(odm.SSDeepHash(copyto="__text__")))
            regkeys = odm.Optional(odm.List(odm.SSDeepHash(copyto="__text__")))

        @odm.model(index=True, store=False)
        class DynamicWindow(odm.Model):
            cls_ids = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            dynamic_classes = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            regkeys = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

        @odm.model(index=True, store=False)
        class DynamicOperatingSystem(odm.Model):
            platform = odm.Optional(odm.List(odm.Platform(copyto="__text__")))
            version = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            processor = odm.Optional(odm.List(odm.Processor(copyto="__text__")))

        autorun_location = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        dos_device = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        mutex = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        registry_key = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        process = odm.Optional(odm.Compound(DynamicProcess))
        signature = odm.Optional(odm.Compound(DynamicSignature))
        ssdeep = odm.Optional(odm.Compound(DynamicSSDeep))
        window = odm.Optional(odm.Compound(DynamicWindow))
        operating_system = odm.Optional(odm.Compound(DynamicOperatingSystem))

    @odm.model(index=True, store=False)
    class Info(odm.Model):
        phone_number = odm.Optional(odm.List(odm.PhoneNumber(copyto="__text__")))

    @odm.model(index=True, store=False)
    class File(odm.Model):
        @odm.model(index=True, store=False)
        class FileDate(odm.Model):
            creation = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            last_modified = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

        @odm.model(index=True, store=False)
        class FileName(odm.Model):
            anomaly = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            extracted = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

        @odm.model(index=True, store=False)
        class FileStrings(odm.Model):
            api = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            blacklisted = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            decoded = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            extracted = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

        @odm.model(index=True, store=False)
        class FileAPK(odm.Model):
            @odm.model(index=True, store=False)
            class FileAPKApp(odm.Model):
                label = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                version = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            @odm.model(index=True, store=False)
            class FileAPKSDK(odm.Model):
                min = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                target = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            activity = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            app = odm.Optional(odm.Compound(FileAPKApp))
            feature = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            locale = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            permission = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            pkg_name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            provides_component = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            sdk = odm.Optional(odm.Compound(FileAPKSDK))
            used_library = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

        @odm.model(index=True, store=False)
        class FileJAR(odm.Model):
            main_class = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            main_package = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

        @odm.model(index=True, store=False)
        class FileIMG(odm.Model):
            @odm.model(index=True, store=False)
            class FileIMGExiftool(odm.Model):
                creator_tool = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                derived_document_id = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                document_id = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                instance_id = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                toolkit = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            exif_tool = odm.Optional(odm.Compound(FileIMGExiftool))
            mega_pixels = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            mode = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            size = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            sorted_metadata_hash = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

        @odm.model(index=True, store=False)
        class FileOLE(odm.Model):
            @odm.model(index=True, store=False)
            class FileOLEMacro(odm.Model):
                sha256 = odm.Optional(odm.List(odm.SHA256(copyto="__text__")))
                suspicious_string = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            @odm.model(index=True, store=False)
            class FileOLESummary(odm.Model):
                author = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                codepage = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                comment = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                company = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                create_time = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                last_printed = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                last_saved_by = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                last_saved_time = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                manager = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                subject = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                title = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            macro = odm.Optional(odm.Compound(FileOLEMacro))
            summary = odm.Optional(odm.Compound(FileOLESummary))
            clsid = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            dde_link = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            fib_timestamp = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

        @odm.model(index=True, store=False)
        class FilePE(odm.Model):
            @odm.model(index=True, store=False)
            class FilePEExports(odm.Model):
                function_name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                module_name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            @odm.model(index=True, store=False)
            class FilePEImports(odm.Model):
                fuzzy = odm.Optional(odm.List(odm.SSDeepHash(copyto="__text__")))
                md5 = odm.Optional(odm.List(odm.MD5(copyto="__text__")))
                sorted_fuzzy = odm.Optional(odm.List(odm.SSDeepHash(copyto="__text__")))
                sorted_sha1 = odm.Optional(odm.List(odm.SHA1(copyto="__text__")))
                suspicious = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            @odm.model(index=True, store=False)
            class FilePELinker(odm.Model):
                timestamp = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            @odm.model(index=True, store=False)
            class FilePEOEP(odm.Model):
                bytes = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                hexdump = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            @odm.model(index=True, store=False)
            class FilePEResources(odm.Model):
                language = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            @odm.model(index=True, store=False)
            class FilePESections(odm.Model):
                hash = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            @odm.model(index=True, store=False)
            class FilePEVersions(odm.Model):
                description = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                filename = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            api_vector = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            exports = odm.Optional(odm.Compound(FilePEExports))
            imports = odm.Optional(odm.Compound(FilePEImports))
            linker = odm.Optional(odm.Compound(FilePELinker))
            oep = odm.Optional(odm.Compound(FilePEOEP))
            pdb_filename = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            resources = odm.Optional(odm.Compound(FilePEResources))
            sections = odm.Optional(odm.Compound(FilePESections))
            versions = odm.Optional(odm.Compound(FilePEVersions))

        @odm.model(index=True, store=False)
        class FilePDF(odm.Model):
            @odm.model(index=True, store=False)
            class FilePDFDate(odm.Model):
                modified = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                pdfx = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                source_modified = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            @odm.model(index=True, store=False)
            class FilePDFJavascript(odm.Model):
                sha1 = odm.Optional(odm.List(odm.SHA1(copyto="__text__")))

            @odm.model(index=True, store=False)
            class FilePDFStats(odm.Model):
                sha1 = odm.Optional(odm.List(odm.SHA1(copyto="__text__")))

            date = odm.Optional(odm.Compound(FilePDFDate))
            javascript = odm.Optional(odm.Compound(FilePDFJavascript))
            stats = odm.Optional(odm.Compound(FilePDFStats))

        @odm.model(index=True, store=False)
        class FilePList(odm.Model):
            @odm.model(index=True, store=False)
            class FilePListBuild(odm.Model):
                machine_os = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            @odm.model(index=True, store=False)
            class FilePListCFBundle(odm.Model):
                @odm.model(index=True, store=False)
                class FilePListCFBundleVersion(odm.Model):
                    long = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                    short = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

                development_region = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                display_name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                executable = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                identifier = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                pkg_type = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                signature = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                url_scheme = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                version = odm.Optional(odm.Compound(FilePListCFBundleVersion))

            @odm.model(index=True, store=False)
            class FilePListDT(odm.Model):
                @odm.model(index=True, store=False)
                class FilePListDTPlatform(odm.Model):
                    build = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                    name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                    version = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

                compiler = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                platform = odm.Optional(odm.Compound(FilePListDTPlatform))

            @odm.model(index=True, store=False)
            class FilePListLS(odm.Model):
                background_only = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                min_system_version = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            @odm.model(index=True, store=False)
            class FilePListNS(odm.Model):
                apple_script_enabled = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                principal_class = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            @odm.model(index=True, store=False)
            class FilePListUI(odm.Model):
                background_modes = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
                requires_persistent_wifi = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            @odm.model(index=True, store=False)
            class FilePListWK(odm.Model):
                app_bundle_identifier = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            installer_url = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            min_os_version = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            requests_open_access = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

            build = odm.Optional(odm.Compound(FilePListBuild))
            cf_bundle = odm.Optional(odm.Compound(FilePListCFBundle))
            dt = odm.Optional(odm.Compound(FilePListDT))
            ls = odm.Optional(odm.Compound(FilePListLS))
            ns = odm.Optional(odm.Compound(FilePListNS))
            ui = odm.Optional(odm.Compound(FilePListUI))
            wk = odm.Optional(odm.Compound(FilePListWK))

        @odm.model(index=True, store=False)
        class FilePowerShell(odm.Model):
            cmdlet = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

        @odm.model(index=True, store=False)
        class FileSWF(odm.Model):
            @odm.model(index=True, store=False)
            class FileSWFHeader(odm.Model):
                @odm.model(index=True, store=False)
                class FileSWFHeaderFrame(odm.Model):
                    count = odm.Optional(odm.List(odm.Integer()))
                    rate = odm.Optional(odm.List(odm.Keyword()))
                    size = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

                frame = odm.Optional(odm.Compound(FileSWFHeaderFrame))
                version = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            header = odm.Optional(odm.Compound(FileSWFHeader))
            tags_ssdeep = odm.Optional(odm.List(odm.SSDeepHash(copyto="__text__")))

        behavior = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        compiler = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        config = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        date = odm.Optional(odm.Compound(FileDate))
        lib = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        name = odm.Optional(odm.Compound(FileName))
        path = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        rule = odm.Optional(odm.Mapping(odm.List(odm.Keyword(copyto="__text__"))))
        string = odm.Optional(odm.Compound(FileStrings))
        apk = odm.Optional(odm.Compound(FileAPK))
        jar = odm.Optional(odm.Compound(FileJAR))
        img = odm.Optional(odm.Compound(FileIMG))
        ole = odm.Optional(odm.Compound(FileOLE))
        pe = odm.Optional(odm.Compound(FilePE))
        pdf = odm.Optional(odm.Compound(FilePDF))
        plist = odm.Optional(odm.Compound(FilePList))
        powershell = odm.Optional(odm.Compound(FilePowerShell))
        swf = odm.Optional(odm.Compound(FileSWF))

    @odm.model(index=True, store=False)
    class Network(odm.Model):
        @odm.model(index=True, store=False)
        class NetworkIOCs(odm.Model):
            domain = odm.Optional(odm.List(odm.Domain(copyto="__text__")))
            ip = odm.Optional(odm.List(odm.IP(copyto="__text__")))
            uri = odm.Optional(odm.List(odm.URI(copyto="__text__")))
            uri_path = odm.Optional(odm.List(odm.URIPath(copyto="__text__")))

        @odm.model(index=True, store=False)
        class NetworkEmail(odm.Model):
            address = odm.Optional(odm.List(odm.Email(copyto="__text__")))
            date = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            subject = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            msg_id = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

        @odm.model(index=True, store=False)
        class NetworkSignature(odm.Model):
            signature_id = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            message = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

        @odm.model(index=True, store=False)
        class NetworkTLS(odm.Model):
            ja3_hash = odm.Optional(odm.List(odm.MD5(copyto="__text__")))
            ja3_string = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
            sni = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

        attack = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        dynamic = odm.Optional(odm.Compound(NetworkIOCs))
        email = odm.Optional(odm.Compound(NetworkEmail))
        mac_address = odm.Optional(odm.List(odm.MAC(copyto="__text__")))
        port = odm.Optional(odm.List(odm.Integer()))
        protocol = odm.Optional(odm.List(odm.Keyword()))
        signature = odm.Optional(odm.Compound(NetworkSignature))
        static = odm.Optional(odm.Compound(NetworkIOCs))
        tls = odm.Optional(odm.Compound(NetworkTLS))
        user_agent = odm.Optional(odm.List(odm.Keyword()))

    @odm.model(index=True, store=False)
    class Technique(odm.Model):
        comms_routine = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        config = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        crypto = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        keylogger = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        macro = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        masking_algo = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        obfuscation = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        packer = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        persistence = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        shellcode = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
        string = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))

    attribution = odm.Optional(odm.Compound(Attribution))
    av = odm.Optional(odm.Compound(AV))
    cert = odm.Optional(odm.Compound(Cert))
    dynamic = odm.Optional(odm.Compound(Dynamic))
    info = odm.Optional(odm.Compound(Info))
    file = odm.Optional(odm.Compound(File))
    network = odm.Optional(odm.Compound(Network))
    source = odm.Optional(odm.List(odm.Keyword(copyto="__text__")))
    technique = odm.Optional(odm.Compound(Technique))
